'use strict';

var spawn = require('child_process').spawn;
var LEP = module.exports;
var dopts = {}; // single shared empty object

LEP.liveServer = "https://acme-v01.api.letsencrypt.org/directory";
LEP.stagingServer = "https://acme-staging.api.letsencrypt.org/directory";
LEP.subcommands = [
  'run'
, 'certonly'
, 'install'
, 'revoke'
, 'rollback'
, 'config_changes'
, 'plugins'
];
LEP.keys = require('./cached-opts');
LEP.parseOptions = function (text) {
  var options = {};
  var re = /--([a-z0-9\-]+)/g;
  var m;

  function uc(match, c) {
    return c.toUpperCase();
  }

  while ((m = re.exec(text))) {
    var key = m[1].replace(/-([a-z0-9])/g, uc);

    options[key] = true;
  }

  return options;
};

LEP.objToArr = function (keys, args) {
  var copy = {};
  var arr = [];

  Object.keys(args).forEach(function (key) {
    var val = args[key];

    if (!val && 0 !== val) {
      // non-zero value which is false, null, or otherwise falsey
      // falsey values should not be passed
      return;
    }

    if (-1 === keys.indexOf(key)) {
      // key is not recognized by the python client
      return;
    }

    if (Array.isArray(val)) {
      copy[key] = args[key].join(',');
    } else {
      copy[key] = args[key];
    }
  });

  Object.keys(copy).forEach(function (key) {
    if ('tlsSni01Port' === key) {
      arr.push('--tls-sni-01-port');
    }
    else if ('http01Port' === key) {
      arr.push('--http-01-port');
    }
    else {
      arr.push('--' + key.replace(/([A-Z])/g, '-$1').toLowerCase());
    }

    if (true !== args[key]) {
      // value is truthy, but not true (and falsies were weeded out above)
      arr.push(args[key]);
    }
  });

  return arr;
};

LEP.opts = function (lebinpath, cb, opts) {
  LEP.exec(lebinpath, null, ['--help', 'all'], function (err, text) {
    if (err) {
      cb(err);
      return;
    }

    cb(null, Object.keys(LEP.parseOptions(text)));
  }, opts);
};

LEP.register = function (lebinpath, subcommand, args, cb, opts) {
  var err;
  if (!args.agreeTos) {
    err = new Error("you did not agree to the terms of service");
    err.code = "INVALID_TERMS";
    cb(err);
    return;
  }

  if ('server' in args && !args.server) {
    err = new Error("Invalid falsey 'server' argument. Specify a string or don't specify it at all.");
    err.code = "INVALID_TERMS";
    cb(err);
  }

  // TODO create and watch the directory for challenge callback
  if (args.challengeCallback) {
    err = new Error("challengeCallback not yet supported");
    cb(err);
    return;
    // args.challengeCallback = false
  }

  if ('configChanges' === subcommand) {
    subcommand = 'config_changes';
  }

  var cliargs = LEP.objToArr(LEP.keys, args);
  return LEP.exec(lebinpath, subcommand, cliargs, cb, opts);
};

LEP.exec = function (lebinpath, subcommand, args, cb, opts) {
  var err;

  if (subcommand) {
    if (-1 === LEP.subcommands.indexOf(subcommand)) {
      cb(new Error("invalid subcommand '" + subcommand + "'"));
      return;
    }
    args.unshift(subcommand);
  }

  if (!opts) {
    opts = dopts;
  }
  if (opts.debug) {
    console.log('[LEP]: letsencrypt-python');
    console.log(lebinpath, args.join(' '));
  }

  var le = spawn(lebinpath, args, { stdio: ['ignore', 'pipe', 'pipe'] });
  var text = '';
  var errtext = '';

  le.on('error', function (error) {
    if (opts.debug) {
      console.error('[LEP]:', error);
    }
    err = error;
  });

  le.stdout.on('data', function (chunk) {
    var str = chunk.toString('utf8');
    if (opts.debug) {
      console.log('[LEP]:');
      console.log(str);
    }
    text += str;
  });

  le.stderr.on('data', function (chunk) {
    var str = chunk.toString('utf8');
    if (opts.debug) {
      console.error('[LEP]:');
      console.error(str);
    }
    errtext += str;
  });

  le.on('close', function (code, signal) {
    if (opts.debug) {
      console.log('[LEP]: close', code, signal);
    }
    if (err) {
      cb(err);
      return;
    }

    if (0 !== code) {
      err = new Error(errtext);
      err.code = code;
      err.signal = signal;
      cb(err);
      return;
    }

    cb(null, text);
  });
};

LEP.create = function (lebinpath, opts) {
  var letsencrypt = {};

  if (!opts) {
    opts = dopts;
  }

  Object.keys(LEP).forEach(function (key) {
    letsencrypt[key] = LEP[key];
  });

  Object.keys(LEP.subcommands).forEach(function (subcommand) {
    if ('config_changes' === subcommand) {
      subcommand = 'configChanges';
    }

    letsencrypt[subcommand] = function (args, cb) {
      letsencrypt.register(subcommand, args, cb);
    };
  });

  letsencrypt.register = function (subcommand, args, cb) {
    LEP.register(lebinpath, subcommand, args, cb, opts);
  };

  return letsencrypt;
};
